
/**
 * @fileoverview
 * Module with functions for parsing atom activity feeds and for serializing
 * objects to the same.
 *
 * Doesn't support implied actions from the spec (yet)
 * 
 */

function ActivityStreamParser()
{
    // There need ever only be one instance since there is no state! this makes
    // it a singleton.
    if (ActivityStreamParser.instance) {
        return ActivityStreamParser.instance;
    }
    else
    {
        ActivityStreamParser.instance = this;
    }

    this.verbMap = {
        "http://activitystrea.ms/schema/1.0/post": "post", 
        "http://activitystrea.ms/schema/1.0/share": "share",
        "http://activitystrea.ms/schema/1.0/save": "save",
        "http://activitystrea.ms/schema/1.0/favorite": "favorite",
        "http://activitystrea.ms/schema/1.0/view": "view",
        "http://activitystrea.ms/schema/1.0/follow": "follow",
        "http://activitystrea.ms/schema/1.0/unfollow": "unfollow",
        "http://activitystrea.ms/schema/1.0/make-friend": "make-friend",
        "http://activitystrea.ms/schema/1.0/join": "join",
        "http://activitystrea.ms/schema/1.0/leave": "leave",
        "http://activitystrea.ms/schema/1.0/invite": "invite"
    }

    this.objectMap = {
         "http://activitystrea.ms/schema/1.0/article": "article",
         "http://activitystrea.ms/schema/1.0/blog-entry": "blog-entry",
         "http://activitystrea.ms/schema/1.0/note": "note",
         "http://activitystrea.ms/schema/1.0/file": "file",
         "http://activitystrea.ms/schema/1.0/image": "image",
         "http://activitystrea.ms/schema/1.0/photo": "photo",
         "http://activitystrea.ms/schema/1.0/photo-album": "photo-album",
         "http://activitystrea.ms/schema/1.0/playlist": "playlist",
         "http://activitystrea.ms/schema/1.0/video": "video",
         "http://activitystrea.ms/schema/1.0/tv-episode": "tv-episode",
         "http://activitystrea.ms/schema/1.0/movie": "movie",
         "http://activitystrea.ms/schema/1.0/audio": "audio",
         "http://activitystrea.ms/schema/1.0/bookmark": "bookmark",
         "http://activitystrea.ms/schema/1.0/person": "person",
         "http://activitystrea.ms/schema/1.0/group": "group",
         "http://activitystrea.ms/schema/1.0/place": "place",

        // This URI is not in the spec, but it's what facebook uses, in spite
        // of the spec saying: "The "note" Object type represents short notes or
        // status updates. This Object type is intended for use in
        // "micro-blogging" and in systems where users are invited to publish a
        // timestamped "status"."
         "http://activitystrea.ms/schema/1.0/status": "note"

    }

    /**
     * Parse an xml document object into javascript objects.
     */
    this.parseDocument = function(doc)
    {
        if (! doc || ! doc.documentElement ||
            ! doc.documentElement.nodeName=="feed")
        {
            // something was wrong with the feed!
            return null;
        }

        return this._doParse(doc.documentElement);
    }

    /**
     * Parse a string containing an xml document into javascript objects.
     */
    this.parseString = function(str)
    {
        var parser = new DOMParser();
        var doc = parser.parseFromString(str, "text/xml");
        return this.parseDocument(doc);
    }

    this._doParse = function(node)
    {
        var meta = this._parseMeta(node);
        var entries = this._parseEntries(node);
        meta.entries = entries;
        return meta;
    }

    /**
     * Parse all of the bits that are not <entry> elements
     */
    this._parseMeta = function(node)
    {
        var ret = {};
        for (var n=0, element; element=node.childNodes[n]; n++)
        {
            switch (element.nodeName) {
                case "id":
                case "title":
                    this._parseLiteral(element, ret);
                    break;
                case "published":
                case "updated":
                    this._parseTimestamp(element, ret);
                    break;
                case "link":
                    this._parseLink(element, ret);
                    break;
                default:
                    //opera.postError("skipping " + element.nodeName)
            }
        }
        return ret;
    }

    /**
     * Strips whitespace!
     */
    this._parseLiteral = function(element, ret)
    {
        ret[element.nodeName] = element.textContent.replace(/^\s*|\s*$/g, "");
    }

    this._parseTimestamp = function(element, ret)
    {
        var d = this._parseRfcDate(element.textContent.replace(/^\s*|\s*$/g, ""));
        ret[element.nodeName] = d;
    }

    this._parseLink = function(element, ret)
    {
        var link = {};

        var href = element.getAttribute("href");
        if (href)
        {
            link.href = href;
        }
        else
        {
            // no href, so the link tag is invalid
            return;
        }

        var rel = element.getAttribute("rel");
        link.rel = rel || "alternate"; // "alternate" is default according to atom spec

        var type = element.getAttribute("type");
        if (type)
        {
            link.type = rel;
        }
        
        if (ret.links) {
            ret.links.push(link);
        }
        else
        {
            ret.links = [link];
        }
    }

    /**
     * Parse all <entry> elements.
     */
    this._parseEntries = function(node)
    {
        var ret = [];
        for (var n=0, element; element=node.childNodes[n]; n++)
        {
            if (element.nodeName == "entry")
            {
                this._parseEntry(element, ret);
            }
        }
        return ret;
    }

    this._parseEntry = function(node, list)
    {
        var ret = {};
        for (var n=0, element; element=node.childNodes[n]; n++)
        {
            switch (element.nodeName) {
                case "#text": // ignore text nodes
                    break;
                case "id":
                case "title":
                case "content":
                    this._parseLiteral(element, ret);
                    break;
                case "published":
                case "updated":
                    this._parseTimestamp(element, ret);
                    break;
                case "link":
                    this._parseLink(element, ret);
                    break;
                case "activity:verb":
                    this._parseVerb(element, ret);
                    break;
                case "activity:object":
                    this._parseObject(element, ret);
                    break;
                case "author":
                    this._parseAuthor(element, ret);
                    break;
                default:
                    //opera.postError("skipping " + element.nodeName)
            }
        }

        // fixme: check if there is no verb. If there isn't do an implied parse
        list.push(ret);
    }

    this._parseAuthor = function(node, ret)
    {
        var author = {};
        for (var n=0, element; element=node.childNodes[n]; n++)
        {
            if (element.nodeName=="#text") {continue}
            this._parseLiteral(element, author);
        }
        ret.author = author;
    }

    this._parseVerb = function(node, ret)
    {
        var uri = node.textContent.replace(/^\s*|\s*$/g, "");

        // hackhack: facebook does not use the identifiers from the spec, as
        // they sometimes append a slash at the end. We try to compensate.
        if (! (uri in this.verbMap) && uri.charAt(uri.length-1)=="/") {
            uri = uri.slice(0, -1);
        }

        ret["verb"] = uri in this.verbMap ? this.verbMap[uri] : "unknown";
    }

    this._parseObject = function(node, ret)
    {
        var data = {};
        for (var n=0, element; element=node.childNodes[n]; n++)
        {
            switch (element.nodeName)
            {
                case "#text": // ignore text nodes
                    break;
                case "link":
                    this._parseLink(element, ret);
                    break;
                case "published":
                case "updated":
                    this._parseTimestamp(element, data);
                    break;
                case "activity:object-type":
                    this._parseObjectType(element, data);
                    break;
                default:
                    this._parseLiteral(element, data);
                    break;
            }
        }
        ret.object = data;
    }

    this._parseObjectType = function(node, ret)
    {
        var uri = node.textContent.replace(/^\s*|\s*$/g, "");
        
        // hackhack: facebook does not use the identifiers from the spec, as
        // they sometimes append a slash at the end. We try to compensate.

        if (! (uri in this.objectMap) && uri.charAt(uri.length-1)=="/") {
            uri = uri.slice(0, -1);
        }
        ret["objectType"] = uri in this.objectMap ? this.objectMap[uri] : null;
        ret["object-type"] = ret["objectType"];
        // duplicate the above so there's a version that works with object
        // syntax and one that uses the exact name from the xml.
    }

    this._parseImpliedActivity = function(node)
    {

    }
    
    /**
     * Parse a date/time in rfc 3339 format into a js date object
     * This code is taken from the widgetize atom parser!
     */
    this._parseRfcDate = function(str) {
        var rfc3339 = /(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})(\.\d+)?(Z|[+-]\d{2}:\d{2})/i;
        var res, hoff, moff; 
        if (res = rfc3339.exec(str)) {
            if (res[8] == 'Z') {
                hoff = 0;
                moff = 0;
            } else {
                var s = res[8].split(':');
                hoff = + s[0];
                moff = + s[1];
            }
            var ms = res[7] ? res[7] * Math.pow( 10, -res[7].length ) : 0;
            return new Date( Date.UTC(res[1], res[2] - 1, res[3], res[4] - hoff, res[5] - moff, res[6], ms ));
        } else {
            return null;    
        }
    }
}
