/**
 * Copyright (c) 2009, Opera Software ASA
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Opera Software ASA nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY OPERA SOFTWARE ASA AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL OPERA SOFTWARE ASA AND CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */


/**
 * @fileoverview
 * Yusef's Communication Manager plugin. AKA comms
 *
 *
 * @author Gautam Chandna, Opera Software ASA (gautamc@opera.com)
 * @version 1.0
 */

Yusef.addPlugin
(
    'comm',
    function CommunicationManager()
    {
        var that            = this;
        var undefined;

        /**
         *  @private
         *  @constant
         */
        this._config = {};
        this._config.pluginName             = 'YUSEF\'s Communication Manager PLUGIN';
        this._config.messageBoxSize         = 100;
        this._config.offlineTimeout         = 30 * 1000;
        this._config.messageSpeed           = 3 * 1000; //between 3 messages
        this._config.maxWarnings            = 4;
        this._config.banTime                = 30 * 60 * 1000;
        
        /**
         *  @private
         */
        this._cliffs = {};
        
        /**
         *  Create a cliff where connections can hang.
         *
         *  @public
         *  @param  {Number}    uid   a unique id for this cliff
         */
        this.createCliff = function( uid )
        {
            if(uid===undefined) return false;
            if(this._cliffs[ uid ]) return false;
            this._cliffs[ uid ] =
            {
                'cliffId'         : uid,
                'banned'          : {},
                'userData'        : {},
                'hangers'         : {},
                'allMessages'     : [],
                'newMessage'      : function( broadcastData ){
                        var newMessage = {
                            "name"        : decodeURI(broadcastData.name),
                            "currentPath" : broadcastData.currentPath,
                            "appName"     : broadcastData.appName,
                            "favicon"     : broadcastData.favicon,
                            "message"     : broadcastData.message,
                            "type"        : broadcastData.type,
                            "format"         : broadcastData.format,
                            "timestamp"   : new Date()
                        }
                        this.allMessages.push(newMessage);
                        if(that._config.messageBoxSize < this.allMessages.length)
                        {
                            this.allMessages.splice( 0, 1 );
                        }
                        return newMessage;
                    },
                'getMessagesSince' : function( connection )
                    {
                        if(connection.messageIndex >= this.allMessages.length)
                        {
                            connection.messageIndex--;
                        }
                        return {"messages" : this.allMessages.slice(connection.messageIndex)};
                    },
                'messageReceived' : function()
                    {
                        return "Please overload this.";
                    },
                'formatData'      : function( newData )
                    {
                        return this.xmlData( newData );
                    },
                'xmlData'         : function( newData )
                    {
                        var data = '<?xml version="1.0" encoding="UTF-8"?>';
                        data += "<items>";
                        for(var i=0; message=newData.messages[i]; i++)
                        {
                            data += "<item>" +
                              " <name>"        + message.name        + "</name>" +
                              " <appName>"     + message.appName     + "</appName>" +
                              " <favicon>"     + message.favicon     + "</favicon>" +
                              " <currentPath>" + message.currentPath + "</currentPath>" +
                              " <timestamp>"   + message.timestamp   + "</timestamp>" +
                              " <message>"     + message.message     + "</message>" +
                              " <type>"        + message.type        + "</type>" +
                              " <format>"      + message.format      + "</format>" +
                              "</item>";
                        }
                        data += "</items>";
                        return data;
                    },
                'updateAllCliffHangers' : function( )
                    {
                        for(key in this.hangers)
                        {
                            this.updateCliffHanger(this.hangers[key]);
                        }
                    },
                'updateCliffHanger' : function( connection )
                    {
                        if( !connection.closed )
                        {
                            var newData = this.getMessagesSince( connection );
                            switch(this.userData[connection.session.id].format)
                            {
                                case "xml":
                                    var data = this.xmlData( newData )||"";
                                    break;
                                default:
                                    var data = this.formatData( newData )||"";
                            }
                            connection.response.setStatusCode( 200 );
                            connection.response.write( data );
                            connection.pending = false;
                            connection.response.close();
                        } else {
                            //This connection was already closed.
                        }
                        this.removeHanger[ connection.hangerId ];
                    },
                'removeHangersBySSID' : function( ssid )
                    {
                        for(hangerId in this.hangers)
                        {
                            if(this.hangers[hangerId].ssid == ssid)
                            {
                                this.removeHanger( hangerId );
                            }
                        }
                    },
                'removeClosedHangers' : function()
                    {
                        for(hangerId in this.hangers)
                        {
                            if(this.hangers[hangerId].closed)
                            {
                                this.removeHanger( hangerId );
                            }
                        }
                    },
                'removeHanger' : function( hangerId )
                    {
                        delete this.hangers[ hangerId ];
                    },
                'removeUserData' : function( ssid )
                    {
                        delete this.userData[ ssid ];
                    },
                'banSSID' : function( ssid )
                    {
                        this.banned[ssid] = new Date();
                    },
                'allowSSID' : function( ssid )
                    {
                        delete this.banned[ssid];
                    },
                'syncCliffWith' : function( urlHang )
                    {
                        new _getRequest( 
                            urlHang, 
                            "unite-format=xml" + 
                            "&unite-name=" + opera.io.webserver.userName +
                            "&unite-cliffId=" + this.cliffId +
                            "&unite-newConnection=true",
                            true,
                            this.cliffId
                        );
                        
                    }
            }
            return this._cliffs[ uid ];
        }
        this.removeCliff = function( uid )
        {
            if(uid===undefined) return false;
            if(this._cliffs[ uid ])
            {
                delete this._cliffs[ uid ];
                return true;
            } else {
                return false;
            }
        }
        this._addCliffHanger = function( connection, userData )
        {
            if( userData.name.length < 3 )
            {
                return '<error type="general">'+_('Please use a longer name.')+'</error>';
            }
            if(userData.name == opera.io.webserver.userName && !connection.isLocal)
            {
                return '<error type="general">'+_('The owner\'s name is reserved.')+'</error>';
            }
            connection.response.setResponseHeader("Content-Type", "text/html");
            connection.response.setResponseHeader("Transfer-Encoding", "chunked");
            connection.response.setResponseHeader("Cache-Control", "no-cache, must-revalidate");
            var cliff = this._cliffs[ userData.cliffId ]||false;
            if(cliff)
            {
                if(connection.session.id in cliff.banned)
                {
                  return '<error type="general">'+_('This session has been banned for 30 minutes because you were too fast.')+'</error>';
                }
                
                if(!cliff.userData[ connection.session.id ])
                {
                    var broadcastData = {
                        "name"        : userData.name,
                        "appName"     : userData.appName,
                        "currentPath" : userData.currentPath,
                        "message"     : _("{name} has joined.", {name: decodeURI(userData.name)}),
                        "format"      : userData.format,
                        "type"        : "server"
                    }
                    if( !(connection.isOwner || connection.isLocal) )
                    {
                        cliff.newMessage( broadcastData );
                        cliff.updateAllCliffHangers();
                        widget.showNotification(_("{name} has joined.", {name: decodeURI(userData.name)}));
                    }
                    if( connection.isOwner )
                    {
                        cliff.newMessage( broadcastData );
                        cliff.updateAllCliffHangers();
                    }
                } else {
                    userData.broadcastAttempt = cliff.userData[ connection.session.id ].broadcastAttempt;
                    userData.warnings = cliff.userData[ connection.session.id ].warnings;
                }
                cliff.userData[ connection.session.id ] = userData;
                cliff.hangers[ connection.hangerId ] = connection;
                cliff.hangers[ connection.hangerId ].pending = true;
                
                if( connection.newConnection )
                {
                    connection.messageIndex = 0;
                    cliff.updateCliffHanger( connection );
                } else {
                    connection.messageIndex = cliff.allMessages.length;
                }
                return true;
            } else {
                return '<error type="general">'+_('Unable to connect to the Shoutbox.')+'</error>';
            }
        }
        function _getRequest( _url, _vars, _reconnect, _cliffId )
        {   
            function _readyState()
            {
                if( this.readyState!==4 )	return false;
                if( this.status==200 )
                {
                    var xml = this.responseXML;
                    if( xml )
                    {
                        var error = (xml.getElementsByTagName('error')[0]||{textContent:''}).textContent;
                        if(error){
                            //_getErrorElement.textContent = error;
                        } else {
                            var messages = xml.getElementsByTagName("item");
                            
                            
                            for(var i=0; message = messages[i]; i++)
                            {
                                var broadcastData = {};
                                broadcastData.name = (message.getElementsByTagName('name')[0]||{textContent:''}).textContent;
                                broadcastData.appName = (message.getElementsByTagName('appName')[0]||{textContent:''}).textContent;
                                broadcastData.favicon = (message.getElementsByTagName('favicon')[0]||{textContent:''}).textContent;
                                broadcastData.currentPath = (message.getElementsByTagName('currentPath')[0]||{textContent:''}).textContent;
                                broadcastData.timestamp = (message.getElementsByTagName('timestamp')[0]||{textContent:''}).textContent;
                                broadcastData.message = (message.getElementsByTagName('message')[0]||{textContent:''}).textContent;
                                broadcastData.type = (message.getElementsByTagName('type')[0]||{textContent:''}).textContent;
                                broadcastData.format = (message.getElementsByTagName('format')[0]||{textContent:''}).textContent;
                                that._cliffs[_cliffId].newMessage( broadcastData );
                            }
                            that._cliffs[_cliffId].updateAllCliffHangers();
                        }
                        if(_reconnect)
                        {
                            _vars = _vars.replace(/&unite-newConnection=true/gi, "");
                            new _getRequest( _url, _vars, _reconnect, _cliffId );
                        }
                    }
                } else {
                    //opera.postError("ERROR " + this.status );
                    //_getErrorElement.innerHTML += "There was an error, please refresh this page.\n";
                }
            }
            
            var xhr = new XMLHttpRequest();
            _vars = _vars||"";
            xhr.open( 'GET', _url + "?" + _vars, true );
            xhr.onreadystatechange = _readyState;
            xhr.send( );
        }
        /**
         *  Initialize method.
         *  Setup the API routes to
         */
        this.initialize = function()
        {
            this.addAPIRoute
            (
                '1.0',
                /cliff/,
                function( connection, matches )
                {
                    if( connection.isOwner==false )
                    {
                        connection.response.setStatusCode("403", "Forbidden");
                        return '<error type="access">'+_('Cliff method only available to the application owner')+'</error>';
                    }
                    var cliffId   = (connection.request.bodyItems['unite-cliffId']||[""])[0];
                    var task   = (connection.request.bodyItems['unite-task']||[""])[0];
                    var result;
                    if(        task == "remove" ){
                        result = that.removeCliff(cliffId);
                    } else if( task == "create" ) {
                        result = that.createCliff(cliffId);
                    }
                    if( result===undefined )
                    {
                        return '<error type="general">'+_('There was no task to perform.')+'</error>'
                    } else if( result===false ){
                        return '<error type="general">'+_('Something went wrong.')+'</error>'
                    }
                    return '<success type="general">'+_('Cliff ' + data.cliffId + ' ' + data.task +'ed.')+'</success>'
                }
            );
            this.addAPIRoute
            (
                '1.0',
                /hang/,
                function( connection, matches )
                {
                    connection.hangerId = connection.session.id + "" + connection.id;
                    connection.newConnection = (connection.request.queryItems['unite-newConnection']||[""])[0];
                    var userData = {}
                    userData.cliffId = (connection.request.queryItems['unite-cliffId']||[""])[0];
                    userData.format  = (connection.request.queryItems['unite-format']||[""])[0];
                    userData.name = (connection.request.queryItems['unite-name']||[""])[0];
                    userData.currentPath = (connection.request.queryItems['unite-currentPath']||[""])[0];
                    userData.appName = unescape((connection.request.queryItems['unite-appName']||[""])[0]);
                    userData.favicon = unescape((connection.request.queryItems['unite-favicon']||[""])[0]);
                    userData.broadcastAttempt = [];
                    userData.warnings = 0;
                    var result = that._addCliffHanger( connection, userData );
                    if( result!==true )
                    {
                        return result;
                    }
                }
            );
            this.addAPIRoute
            (
                '1.0',
                /broadcast/,
                function( connection, matches )
                {
                    var broadcastData = {}
                    broadcastData.cliffId = (connection.request.bodyItems['unite-cliffId']||[""])[0];
                    broadcastData.message = unescape((connection.request.bodyItems['unite-message']||[""])[0]);
                    broadcastData.currentPath = unescape((connection.request.bodyItems['unite-currentPath']||[""])[0]);
                    broadcastData.appName = unescape((connection.request.bodyItems['unite-appName']||[""])[0]);
                    broadcastData.favicon = unescape((connection.request.bodyItems['unite-favicon']||[""])[0]);
                    if(broadcastData.message != "")
                    {
                        cliff = that._cliffs[ broadcastData.cliffId ]||false;
                        if(cliff)
                        {
                            //Default data
                            broadcastData.name    = cliff.userData[connection.session.id].name;
                            broadcastData.type = "client";
                            //Spam control
                            var broadcastAttempt = cliff.userData[connection.session.id].broadcastAttempt;
                            broadcastAttempt.push(new Date());
                            if(broadcastAttempt.length > 4)
                            {
                                broadcastAttempt.splice(0,1);
                            }
                            if(broadcastAttempt.length == 4)
                            {
                                speed = broadcastAttempt[3]-broadcastAttempt[0];
                                if(speed < that._config.messageSpeed && !connection.isOwner)
                                {
                                    var warnings = cliff.userData[connection.session.id].warnings++;
                                    if(warnings > that._config.maxWarnings)
                                    {
                                        broadcastData.message = decodeURI(cliff.userData[connection.session.id].name) + " is typing too fast, and not listening to the warnings.";
                                        broadcastData.type = "server";
                                        cliff.newMessage( broadcastData );
                                        cliff.updateAllCliffHangers();
                                        cliff.removeHangersBySSID(connection.session.id);
                                        cliff.removeUserData(connection.session.id);
                                        cliff.banSSID(connection.session.id);
                                    } else {
                                        cliff.userData[connection.session.id].broadcastAttempt.splice(0,4);
                                        return "<error type='speed'>"+_("Please slow down! Wait a bit.")+"</error>";
                                    }
                                }
                            }
                            
                            cliff.newMessage( broadcastData );
                            cliff.updateAllCliffHangers()
                        }
                        return "<success>"+_("Message broadcasted.")+"</success>"
                    } else {
                        return '<error type="content">'+_('The message was blank.')+'</error>'
                    }
                }
            );
            this.addAPIRoute
            (
                '1.0',
                /closeAll/,
                function( connection, matches )
                {
                    data = that.getData(connection);
                    if( connection.isOwner==false )
                    {
                        connection.response.setStatusCode("403", "Forbidden");
                        return '<error type="access">'+_('closeAll method only available to the application owner')+'</error>';
                    }
                    
                    var cliff = that._cliffs[ data.cliffId ]||false;
                    for(id in cliff.hangers)
                    {
                        delete cliff.hangers[id];
                    }
                    return "<success>"+_("All connections closed")+"</success>"
                }
            );
            function knuckleBreaker()
            {
                var keepTheseSessions = {};
                for(cliffId in that._cliffs)
                {
                    var timestamp = new Date();
                    for(banId in that._cliffs[cliffId].banned)
                    {
                        if((timestamp-that._cliffs[cliffId].banned[banId])>that._config.banTime)
                        {
                            that._cliffs[cliffId].allowSSID(banId);
                        }
                    }
                    that._cliffs[cliffId].removeClosedHangers();
                    for(safeHangerId in that._cliffs[cliffId].hangers)
                    {
                        keepTheseSessions[that._cliffs[cliffId].hangers[safeHangerId].ssid] = true;
                    }
                    for(ssid in that._cliffs[cliffId].userData)
                    {
                        if(!keepTheseSessions[ssid] === true)
                        {
                            var cliff = that._cliffs[cliffId];
                            var broadcastData = {}
                            broadcastData.cliffId = cliffId;
                            broadcastData.message = _("{name} has left {appName}.", {name: decodeURI(cliff.userData[ssid].name), appName: cliff.userData[ssid].appName});
                            broadcastData.type = "server";
                            broadcastData.name    = cliff.userData[ssid].name;
                            broadcastData.currentPath = cliff.userData[ssid].currentPath;
                            broadcastData.appName = cliff.userData[ssid].appName;
                            broadcastData.favicon = cliff.userData[ssid].favicon;
                            if( cliff.userData[ssid].name != opera.io.webserver.userName )
                            {
                                cliff.newMessage( broadcastData );
                                cliff.updateAllCliffHangers()
                            }
                            cliff.removeUserData( ssid );
                        }
                    }
                }
            }
            this.timer = window.setInterval( knuckleBreaker, that._config.offlineTimeout );
        }
    },
    'initialize'
);